@@ -19,8 +19,8 @@ impl Slice {
         Slice { axis, start: start.to_dim(), end: end.to_dim() }
     }
 
-    pub fn suffix(&self) -> String {
-        format!("axis{}_{}_{}", self.axis, self.start, self.end)
+    pub fn suffix(&self, name: &str) -> String {
+        format!("{}.axis{}_{}_{}", name, self.axis, self.start, self.end)
     }
 }
 
@@ -52,9 +52,12 @@ impl EvalOp for Slice {
 
     fn eval(&self, mut inputs: TVec<Arc<Tensor>>) -> TractResult<TVec<Arc<Tensor>>> {
         let input = args_1!(inputs);
+        let start = self.start.to_usize()?;
+        let end = self.end.to_usize()?;
+        if end > input.shape()[self.axis] || start > end {
+            bail!("Invalid range {}..{} for slicing {:?} on axis {}", start, end, input, self.axis);
+        }
         unsafe {
-            let start = self.start.to_usize()?;
-            let end = self.end.to_usize()?;
             let mut shape: TVec<_> = input.shape().into();
             shape[self.axis] = end - start;
             let mut tensor = Tensor::uninitialized_dt(input.datum_type(), &shape)?;
@@ -126,7 +129,7 @@ impl TypedOp for Slice {
             model,
             prec,
             &mut patch,
-            &self.suffix(),
+            &self.suffix(&node.name),
             node.inputs[0].slot,
             self.axis,
             start,
@@ -159,14 +162,15 @@ impl TypedOp for Slice {
     ) -> TractResult<Option<OutletId>> {
         let prec = model.node(node.inputs[0].node);
         if axis != self.axis {
-            let suffix = suffix.to_string() + "." + &self.suffix();
+            let suffix = self.suffix(&node.name) + "." + suffix;
             return prec
                 .op()
                 .as_typed()
                 .unwrap()
                 .slice_output(model, &prec, patch, &suffix, node.inputs[0].slot, axis, start, end)?
                 .map(|w| {
-                    Ok(patch.wire_node(format!("{}.{}", node.name, &suffix), self.clone(), &[w])?[0])
+                    Ok(patch.wire_node(format!("{}.{}", node.name, &suffix), self.clone(), &[w])?
+                        [0])
                 })
                 .transpose();
         }
