@@ -61,7 +61,12 @@ impl ConvUnary {
         )
     }
 
-    fn kernel_as_packed_as(&self, packer: &Packer, m: usize) -> TractResult<ArrayD<Arc<Tensor>>> {
+    fn kernel_as_packed_as(
+        &self,
+        packer: &Packer,
+        k: usize,
+        m: usize,
+    ) -> TractResult<ArrayD<Arc<Tensor>>> {
         let kernel = self.kernel_as_group_o_ihw()?;
         unsafe {
             let mut packed_as = Array1::from(
@@ -69,7 +74,7 @@ impl ConvUnary {
                     .map(|g| {
                         let mut packed = Tensor::uninitialized_aligned_dt(
                             kernel.datum_type(),
-                            &[packer.len(m)],
+                            &[packer.len(k, m)],
                             packer.alignment(),
                         )?;
                         packer.pack(
@@ -233,7 +238,7 @@ impl ConvUnary {
 
         let mut sum_b = model.wire_node(
             format!("{}.sum_b", name),
-            super::QSumB { n: n.clone(), r: mmm.b_pack(k).panel_width(), k },
+            super::QSumB { n: n.clone(), r: mmm.b_pack().panel_width(), k },
             &[im2col],
         )?[0];
 
@@ -413,33 +418,55 @@ impl ConvUnary {
         Ok(wire[0])
     }
 
-    pub unsafe fn wire_as_direct(
+    pub unsafe fn wire_as_lazy_im2col(
         &self,
         model: &mut TypedModel,
         name: &str,
-        wire: OutletId,
-        input_shape: &[usize],
+        mut wire: OutletId,
     ) -> TractResult<OutletId> {
-        let b_fact = model.outlet_fact(wire)?.clone();
-        let c_dt = crate::ops::matmul::output_type(b_fact.datum_type);
+        let mut b_fact = model.outlet_fact(wire)?.clone();
         let (geo, m, k, n, mmm) = self.compute_geo(&b_fact)?;
-        let geo = geo.to_concrete(input_shape)?;
-        let input_shape: DataShape = self.pool_spec.data_format.shape(input_shape.into())?;
+        let input_shape = b_fact.shape.as_concrete().unwrap().to_vec();
+        let mut geo = geo.to_concrete(&input_shape)?.into_owned();
+        let mut input_shape: DataShape = self.pool_spec.data_format.shape(input_shape.into())?;
+        let padding = self.pool_spec.computed_padding(input_shape.hw_dims());
+        if padding.iter().any(|axis| axis.pad_before != 0 || axis.pad_after != 0) {
+            let mut pads = vec![(0, 0); b_fact.rank()];
+            for (ix, ax) in padding.iter().enumerate() {
+                pads[input_shape.h_axis() + ix] = (ax.pad_before, ax.pad_after);
+            }
+            let op = crate::ops::array::Pad {
+                mode: crate::ops::array::PadMode::Constant(
+                    Tensor::zero_scalar_dt(b_fact.datum_type)?.into_arc_tensor(),
+                ),
+                pads,
+            };
+            wire = model.wire_node(format!("{}.pad", name), op, &[wire])?[0];
+            let valid_pool_spec =
+                PoolSpec { padding: ops::cnn::PaddingSpec::Valid, ..self.pool_spec.clone() };
+            b_fact = model.outlet_fact(wire)?.clone();
+            let concrete_shape = b_fact.shape.as_concrete().unwrap();
+            input_shape = valid_pool_spec.data_format.shape(concrete_shape.into())?;
+            geo = valid_pool_spec
+                .compute_geo(&b_fact.shape)?
+                .to_concrete(&*concrete_shape)?
+                .into_owned();
+        }
+        let c_dt = crate::ops::matmul::output_type(b_fact.datum_type);
         let c_stride = input_shape.c_stride();
-        let data_offsets: Vec<isize> = geo.patch.centers_offsets();
-        let kernel_offsets: Vec<isize> = (0..self.input_channels())
+        let size_of_b = b_fact.datum_type.size_of() as isize;
+        let n_bytes_offsets: Vec<isize> =
+            geo.patch.centers_offsets().into_iter().map(|x| x * size_of_b).collect();
+        let k_bytes_offsets: Vec<isize> = (0..self.input_channels())
             .flat_map(|ici| {
                 geo.patch
                     .standard_layout_data_field
                     .iter()
-                    .map(move |x| x + (ici * c_stride) as isize)
+                    .map(move |x| (x + (ici * c_stride) as isize) * size_of_b)
             })
             .collect();
-        let b_storage = mmm.b_from_data_and_offsets(
-            b_fact.datum_type.size_of(),
-            &kernel_offsets,
-            &data_offsets,
-        );
+        let virtual_input = super::lazy_im2col::LazyIm2colSpec { n_bytes_offsets, k_bytes_offsets };
+        let b_storage = mmm.b_virtual_input(Box::new(virtual_input), k);
         let (mmm_output_shape, c_axis, h_axis) = self.mmm_output_shape(&geo.output_shape)?;
 
         let geometry = MatMulGeometry::Concrete(ConcreteMatMulGeometry {
@@ -503,7 +530,7 @@ impl ConvUnary {
         c_m_axis: usize,
         c_n_axis: usize,
     ) -> TractResult<OutletId> {
-        let kernels = self.kernel_as_packed_as(&mmm.a_pack(k), m)?;
+        let kernels = self.kernel_as_packed_as(&mmm.a_pack(), k, m)?;
         let shape = kernels.shape();
         let mut fused_ops = dispatch_copy!(Self::bias_as_non_linear(mmm.internal_type())(self))?;
         for fo in &mut fused_ops {
@@ -554,7 +581,7 @@ impl ConvUnary {
         let input_fact = model.outlet_fact(node.inputs[0])?;
         let spatial_rank = self.kernel.rank() - 2;
         if let Some(axis) = (0..spatial_rank).find(|&ax| {
-            self.pool_spec.padding.valid_dim(ax)
+            self.pool_spec.padding.valid_dim(ax, self.pool_spec.stride(ax) == 1)
                 && self.pool_spec.stride(ax) > 1
                 && self.pool_spec.dilation(ax) % self.pool_spec.stride(ax) == 0
         }) {
@@ -745,11 +772,11 @@ impl TypedOp for ConvUnary {
         }
         if self.pool_spec.output_channel_override != Some(self.output_channels()) {
             bail!(
-                    "Inconsistent convolution: output channels from pool spec is {:?}, kernel expects {} output channels, {:?}",
-                    self.pool_spec.output_channel_override,
-                    self.output_channels(),
-                    self
-                    );
+                "Inconsistent convolution: output channels from pool spec is {:?}, kernel expects {} output channels, {:?}",
+                self.pool_spec.output_channel_override,
+                self.output_channels(),
+                self
+                );
         }
         if let Some(bias) = &self.bias {
             if bias.len() != self.output_channels() {
@@ -955,6 +982,7 @@ impl TypedOp for ConvUnary {
             let new_op = self.kernel_offset_u8_as_i8(&mut inputs, &mut patch)?.unwrap();
             let wire = patch.wire_node(&node.name, new_op, &inputs)?;
             patch.shunt_outside(model, node.id.into(), wire[0])?;
+            patch.obliterate(node.id)?;
             return Ok(Some(patch));
         }
 
@@ -979,6 +1007,7 @@ impl TypedOp for ConvUnary {
                     &inputs,
                 )?;
                 patch.shunt_outside(model, node.id.into(), wire)?;
+                patch.obliterate(node.id)?;
                 return Ok(Some(patch));
             } else if kernel_spatial_shape.iter().product::<usize>() == 1
                 && (0..spatial_rank)
@@ -1045,12 +1074,13 @@ impl TypedOp for ConvUnary {
                     &[wire],
                 )?[0];
                 patch.shunt_outside(model, OutletId::new(node.id, 0), wire)?;
+                patch.obliterate(node.id)?;
                 return Ok(Some(patch));
             } else if input_fact
                 .shape
                 .as_concrete()
                 .map(|s| {
-                    should_use_direct(
+                    should_use_lazy(
                         &self.pool_spec.data_format.shape(s.into()).unwrap(),
                         &self.pool_spec,
                         self.group,
@@ -1058,17 +1088,11 @@ impl TypedOp for ConvUnary {
                 })
                 .unwrap_or(false)
             {
-                let mut patch = TypedModelPatch::default();
-                let wire = patch.tap_model(model, node.inputs[0])?;
-                let wire = self
-                    .wire_as_direct(
-                        &mut patch,
-                        &*node.name,
-                        wire,
-                        input_fact.shape.as_concrete().unwrap(),
-                    )
-                    .context("in wire_as_direct")?;
+                let mut patch = TypedModelPatch::new("wire_as_lazy_im2col");
+                let mut wire = patch.tap_model(model, node.inputs[0])?;
+                wire = self.wire_as_lazy_im2col(&mut patch, &*node.name, wire)?;
                 patch.shunt_outside(model, OutletId::new(node.id, 0), wire)?;
+                patch.obliterate(node.id)?;
                 return Ok(Some(patch));
             } else if self.group != 1
                 && self.group == self.output_channels()
@@ -1085,6 +1109,7 @@ impl TypedOp for ConvUnary {
                     .wire_as_im2col_pair(&mut patch, &*node.name, wire)
                     .context("in wire_as_im2col_pair")?;
                 patch.shunt_outside(model, OutletId::new(node.id, 0), wire)?;
+                patch.obliterate(node.id)?;
                 return Ok(Some(patch));
             }
         }
@@ -1093,19 +1118,8 @@ impl TypedOp for ConvUnary {
     as_op!();
 }
 
-fn should_use_direct(input_shape: &DataShape, pool_spec: &PoolSpec, group: usize) -> bool {
-    let spatial_rank = input_shape.hw_rank();
-    if group != 1 || !(0..spatial_rank).all(|ax| pool_spec.padding.valid_dim(ax)) {
-        return false;
-    }
-    let direct =
-            // no real rationale here, pure heuristic to force "right" pick in
-            // both hey_snips v3 and v4. just hope this will generalize ok
-            (0..spatial_rank).any(|d| pool_spec.dilation(d) > 1 && pool_spec.kernel_shape[d] > 2) ||
-            // that one kind of make sense, better use direct that generate a huge
-            // im2col matrix (when both kernel and input are big)
-            pool_spec.kernel_shape.iter().product::<usize>() * input_shape.shape.iter().product::<usize>() > 1048576;
-    direct
+fn should_use_lazy(_input_shape: &DataShape, pool_spec: &PoolSpec, group: usize) -> bool {
+    group == 1 && pool_spec.kernel_shape.iter().product::<usize>() > 5
 }
 
 #[allow(non_snake_case)]
@@ -1144,120 +1158,4 @@ mod test {
         let output = op.eval(input).unwrap();
         assert_eq!(&*output[0], &tensor4(&[[[[8i32, 12], [20, 24]]]]));
     }
-
-    #[test]
-    fn conv_vs_direct_arm_ml_kws_cnn_m_0() {
-        let input = NHWC.from_n_c_hw(1, 1, &[49, 10]).unwrap();
-        let pool = {
-            PoolSpec::new(
-                HWC,
-                tvec!(10, 4),
-                PaddingSpec::Valid,
-                Some(tvec!(1, 1)),
-                Some(tvec!(1, 1)),
-                Some(64),
-            )
-        };
-        assert!(!should_use_direct(&input, &pool, 1));
-    }
-
-    #[test]
-    fn conv_vs_direct_arm_ml_kws_cnn_m_1() {
-        let input = NHWC.from_n_c_hw(1, 64, &[40, 7]).unwrap();
-        let pool = {
-            PoolSpec::new(
-                HWC,
-                tvec!(10, 4),
-                PaddingSpec::Valid,
-                Some(tvec!(2, 1)),
-                Some(tvec!(1, 1)),
-                Some(48),
-            )
-        };
-        assert!(should_use_direct(&input, &pool, 1));
-    }
-
-    #[test]
-    fn conv_vs_direct_hey_snips_v31() {
-        use crate::ops::cnn::PaddingSpec;
-        use DataFormat::HWC;
-
-        fn dil_use_direct(size: usize, d: usize) -> bool {
-            let input = HWC.from_n_c_hw(1, 128, &[size]).unwrap();
-            let pool = {
-                PoolSpec::new(
-                    HWC,
-                    tvec!(2),
-                    PaddingSpec::Valid,
-                    Some(tvec!(d)),
-                    Some(tvec!(1)),
-                    Some(64),
-                )
-            };
-            should_use_direct(&input, &pool, 1)
-        }
-        assert!(!dil_use_direct(36, 1));
-        assert!(!dil_use_direct(33, 2));
-        assert!(!dil_use_direct(27, 4));
-        assert!(!dil_use_direct(18, 8));
-    }
-
-    #[test]
-    fn conv_vs_direct_hey_snips_v4() {
-        fn dil_use_direct(d: usize) -> bool {
-            let input = HWC.from_n_c_hw(1, 16, &[8 + 2 * d]).unwrap();
-            let pool = {
-                PoolSpec::new(
-                    HWC,
-                    tvec!(3),
-                    PaddingSpec::Valid,
-                    Some(tvec!(d)),
-                    Some(tvec!(1)),
-                    Some(64),
-                )
-            };
-            should_use_direct(&input, &pool, 1)
-        }
-        assert!(!dil_use_direct(1));
-        assert!(dil_use_direct(2));
-        assert!(dil_use_direct(4));
-        assert!(dil_use_direct(8));
-    }
-
-    #[test]
-    fn conv_vs_direct_am_lda_2M() {
-        let pool = {
-            PoolSpec::new(
-                HWC,
-                tvec!(5),
-                PaddingSpec::Valid,
-                Some(tvec!(1)),
-                Some(tvec!(1)),
-                Some(200),
-            )
-        };
-        let input = HWC.from_n_c_hw(1, 40, &[28]).unwrap();
-        assert!(!should_use_direct(&input, &pool, 1));
-    }
-
-    #[test]
-    fn conv_vs_direct_hey_am_tdnn_2M() {
-        fn use_direct(size: usize, stride: usize) -> bool {
-            let input = HWC.from_n_c_hw(1, 256, &[size]).unwrap();
-            let pool = {
-                PoolSpec::new(
-                    HWC,
-                    tvec!(3),
-                    PaddingSpec::Valid,
-                    Some(tvec!(1)),
-                    Some(tvec!(stride)),
-                    Some(64),
-                )
-            };
-            should_use_direct(&input, &pool, 1)
-        }
-        assert!(!use_direct(26, 1)); // tdnn2
-        assert!(!use_direct(24, 3)); // tdnn3
-        assert!(!use_direct(10, 1)); // tdnn4,5
-    }
 }
