@@ -107,7 +107,7 @@ impl EvalOp for ElementWiseOp {
 
 impl TypedOp for ElementWiseOp {
     fn output_facts(&self, inputs: &[&TypedFact]) -> TractResult<TVec<TypedFact>> {
-        let mut fact = inputs[0].clone();
+        let mut fact = inputs[0].clone().without_value();
         if let Some(dt) = self.0.output_type(fact.datum_type) {
             fact.datum_type = dt;
         }
@@ -178,7 +178,7 @@ macro_rules! element_wise {
         #[derive(Debug, Clone, Educe)]
         #[educe(Hash)]
         pub struct $Op { $( $( $(#[$meta])? pub $var: $var_typ),* )? }
-        $crate::impl_dyn_hash!($Op);
+        tract_data::internal::impl_dyn_hash!($Op);
         impl $crate::ops::element_wise::ElementWiseMiniOp for $Op {
             fn name(&self) -> String {
                 format!("{}{}", self.prefix(), stringify!($Op))
@@ -267,7 +267,7 @@ macro_rules! element_wise_oop {
         #[derive(Debug, Clone, Educe)]
         #[educe(Hash)]
         pub struct $Op { $( $($(#[$meta])? pub $var: $var_typ),* )? }
-        $crate::impl_dyn_hash!($Op);
+        tract_data::internal::impl_dyn_hash!($Op);
         impl $crate::ops::element_wise::ElementWiseMiniOp for $Op {
             fn name(&self) -> String {
                 format!("{}{}", self.prefix(), stringify!($Op))
