@@ -593,7 +593,7 @@ mod tests {
         model.set_output_outlets(&[y])?;
         let result = SimplePlan::new(&model)?.run(tvec!(tensor2(&[[1, 2], [3, 4]])))?;
         assert_eq!(result[0], rctensor2(&[[4, 8], [12, 16]]));
-        let decluttered = model.declutter()?;
+        let decluttered = model.into_decluttered()?;
         let result = SimplePlan::new(&decluttered)?.run(tvec!(tensor2(&[[1, 2], [3, 4]])))?;
         assert_eq!(result[0], rctensor2(&[[4, 8], [12, 16]]));
         let op = decluttered.node(1).op().downcast_ref::<UnaryOp>().unwrap();
@@ -610,7 +610,7 @@ mod tests {
         model.set_output_outlets(&[y])?;
         let result = SimplePlan::new(&model)?.run(tvec!(tensor2(&[[16, 32], [64, 68]])))?;
         assert_eq!(result[0], rctensor2(&[[4, 8], [16, 17]]));
-        let decluttered = model.declutter()?;
+        let decluttered = model.into_decluttered()?;
         let result = SimplePlan::new(&decluttered)?.run(tvec!(tensor2(&[[16, 32], [64, 68]])))?;
         assert_eq!(result[0], rctensor2(&[[4, 8], [16, 17]]));
         let op = decluttered.node(1).op().downcast_ref::<UnaryOp>().unwrap();
