@@ -190,11 +190,11 @@ impl MatMulUnary {
             Array::from_shape_fn(&self.a.shape()[0..self.a.rank() - 2], |a_prefix| unsafe {
                 let mut pa = Tensor::uninitialized_aligned_dt(
                     self.a.datum_type(),
-                    &[mmm.a_pack(k).len(m)],
-                    mmm.a_pack(k).alignment(),
+                    &[mmm.a_pack().len(k, m)],
+                    mmm.a_pack().alignment(),
                 )
                 .unwrap();
-                mmm.a_pack(k).pack(
+                mmm.a_pack().pack(
                     &mut pa.view_mut(),
                     &self.a.view_at_prefix(a_prefix.slice()).unwrap(),
                     !self.a_trans as usize,
@@ -204,11 +204,11 @@ impl MatMulUnary {
             });
         unsafe {
             let mut packed_b_shape: TVec<usize> = b_shape[..b_shape.len() - 2].into();
-            packed_b_shape.push(mmm.b_pack(k).len(n));
+            packed_b_shape.push(mmm.b_pack().len(k, n));
             wire = patch.wire_node(
                 format!("{}.pack", &*node.name),
                 super::MatMatMulPack {
-                    packer: mmm.b_pack(k),
+                    packer: mmm.b_pack(),
                     trans: self.b_trans,
                     output_shape: packed_b_shape,
                 },
@@ -238,6 +238,7 @@ impl MatMulUnary {
                 &[wire],
             )?[0];
             patch.shunt_outside(model, OutletId::new(node.id, 0), wire)?;
+            patch.obliterate(node.id)?;
         }
         Ok(patch)
     }
